import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import layers

class DenseNetwork():
	def __init__(
		self,
		input_size,
		output_size,
		hidden_layers=[],
		name="dense network"
	):
		self.name = name
		self.input_size = input_size
		self.output_size = output_size
		self.hidden_layers = hidden_layers

		self.model = self.build_model()

	def build_model(self):
		input_layer = layers.Input(shape=self.input_size, name=self.name, dtype=tf.float32)
		output_layer = None

		if len(self.hidden_layers) == 0:
			output_layer = layers.Dense(self.output_size, activation='relu')(input_layer)
		else:
			if self.hidden_layers[0] > 0.0 and self.hidden_layers[0] < 1.0:
				x = layers.Dropout(rate=self.hidden_layers[0])(input_layer)
			else:
				x = layers.Dense(self.hidden_layers[0], activation='relu')(input_layer)

			for i in range(1, len(self.hidden_layers)):
				if self.hidden_layers[i] > 0.0 and self.hidden_layers[i] < 1.0:
					x = layers.Dropout(rate=self.hidden_layers[i])(x)
				else:
					x = layers.Dense(self.hidden_layers[i], activation='relu')(x)

			output_layer = layers.Dense(self.output_size, activation='relu')(x)

		model = keras.Model(inputs=input_layer, outputs=output_layer, name=self.name)
		return model

	def get_model(self):
		return self.model




import logging
from typing import List
import chess
import chess.pgn

def filter_out_bullet_games(header: chess.pgn.Headers, debug: bool = False):
	return filter_by_elo(header, white_elo_range=[700, 3000], black_elo_range=[700, 3000], debug=debug) and \
		filter_by_time_control(header, time_range=[2, 30], debug=debug)

def filter_out_low_elo_bullet_games(header: chess.pgn.Headers, debug: bool = False):
	return filter_by_elo(header, white_elo_range=[2600, 4000], black_elo_range=[2600, 4000], debug=debug) and \
		filter_by_time_control(header, time_range=[2, 30], debug=debug)

def filter_by_elo(header: chess.pgn.Headers, white_elo_range, black_elo_range, debug=False):
	logger = logging.getLogger(__name__)
	#headers are often non-standard, try..except!
	discard_game = False
	try:
		if not white_elo_range[0] < header.get("WhiteElo") < white_elo_range[1]:
			discard_game = True
		elif not black_elo_range[0] < header.get("BlackElo") < black_elo_range[1]:
			discard_game = True
	except Exception as e:
		if debug:
			logger.error(f"Exception in filter_by_elo", exec_info=True)
		discard_game = True
	finally:
		return discard_game

def filter_by_time_control(header: chess.pgn.Headers, time_range: List[int], debug: bool = False):
	logger = logging.getLogger(__name__)
	#headers are often non-standard, try..except!
	discard_game = False
	try:
		minutes, increment_seconds = get_time_and_increment_from_time_control_header(header.get("TimeControl"))
		total_time = increment_to_total_time_equivalent(increment_seconds, 40)
		if not time_range[0] < total_time < time_range[1]:
			discard_game = True
	except Exception as e:
		if debug:
			logger.error(f"Exception in filter_by_time_control", exec_info=True)
		discard_game = True
	finally:
		return discard_game
		
def get_time_and_increment_from_time_control_header(time_control: str):
	minute, second = time_control.split("+")
	return int(minute), int(second)

def increment_to_total_time_equivalent(increment: int, moves: int):
	return increment * moves / 60



import logging
from typing import Callable, Union

import chess
import chess.pgn
import numpy as np

from chesspos.preprocessing.board_converter import board_to_bitboard, board_to_tensor

def sparse_positions_to_tensor(game: chess.pgn.Game):
	return single_positions(game, board_to_tensor, heavy_subsampling)

def positions_to_tensor(game: chess.pgn.Game):
	return single_positions(game, board_to_tensor, subsample_opening_positions)

def positions_to_bitboard(game: chess.pgn.Game):
	return single_positions(game, board_to_bitboard, subsample_opening_positions)

def positions_to_tensor_triplets(game: chess.pgn.Game) -> np.ndarray:
	position_encodings = single_positions(game, board_to_tensor, subsample_opening_positions)
	number_positions = position_encodings.shape[0]
	anchor_index = np.random.randint(number_positions-1)
	positive_index = anchor_index + 1
	negative_index = (anchor_index + number_positions // 2) % number_positions
	triplets = position_encodings[[anchor_index, positive_index, negative_index],...]
	return triplets.reshape(1, 3, *position_encodings.shape[1:])

def single_positions(game: chess.pgn.Game, position_encoder: Callable[[chess.Board], np.ndarray],
	sample_move: Callable[[Union[int,chess.Move, None]], bool]) -> np.ndarray:
	logger = logging.getLogger(__name__)
	output_shape = position_encoder(chess.Board()).shape
	output_type = position_encoder(chess.Board()).dtype
	output = np.empty((0, *output_shape), dtype=output_type)

	board = chess.Board()
	for i, move in enumerate(game.mainline_moves()):
		try:
			board.push(move)
		except Exception as e:
			logger.error(f"Exception occurred in position number {i}", exec_info=True)
			return output

		if sample_move(i):
			output = np.append(output, position_encoder(board).reshape(1,*output_shape), axis=0)
	return output

def subsample_opening_positions(ply: int) -> bool:
	selection_probability = min(ply / 40, 1.0)
	return np.random.random() < selection_probability

def heavy_subsampling(ply: int):
	selection_probability = subsample_opening_positions(ply) / 50.0
	return np.random.random() < selection_probability


from functools import wraps
import numpy as np
import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import layers, Model

from chesspos.models.autoencoder import AutoencoderModel

class ResnetAutoencoder(AutoencoderModel):
	@wraps(AutoencoderModel.__init__)
	def __init__(self, **kwargs):
		self._final_conv_shape = None
		self.embedding_size = 256
		super().__init__(**kwargs)

	def _model_helper(self) -> dict:
		encoder_input = layers.Input(shape=(8,8,15,1), dtype=tf.float16)

		conv_1 = layers.Conv3D(16, (2,2,15), activation="relu", padding="same")(encoder_input)
		conv_1 = layers.BatchNormalization()(conv_1)
		conv_2 = layers.Conv3D(32, (2,2,15), activation="relu", padding="same")(conv_1)
		conv_2 = layers.BatchNormalization()(conv_2)
		#conv_3 = layers.Conv3D(64, (2,2,15), activation="relu", padding="same")(conv_2)
		#conv_3 = layers.BatchNormalization()(conv_3)
		#conv_4 = layers.Conv3D(64, (2,2,15), activation="relu", padding="same")(conv_3)
		#conv_4 = layers.BatchNormalization()(conv_4)
		#conv_5 = layers.Conv3D(128, (2,2,15), activation="relu", padding="same")(conv_4)
		#conv_5 = layers.BatchNormalization()(conv_5)
		#conv_6 = layers.Conv3D(256, (2,2,15), activation="relu", padding="same")(conv_5)
		#conv_6 = layers.BatchNormalization()(conv_6)
		#conv_7 = layers.Conv3D(512, (2,2,15), activation="relu", padding="same")(conv_6)
		#conv_7 = layers.BatchNormalization()(conv_7)
		
		self._final_conv_shape = conv_2.shape[1:]
		print("final_conv_shape:", self._final_conv_shape)
		embedding = layers.Flatten()(conv_2)
		embedding = layers.Dense(self.embedding_size, activation="relu")(embedding)

		decoder_input = layers.Input(shape=(self.embedding_size,))
		decoder_dense = layers.Dense(np.prod(self._final_conv_shape), activation="relu")(decoder_input)
		decoder_dense = layers.Reshape(self._final_conv_shape)(decoder_dense)
		decoder_conv_1 = layers.Conv3DTranspose(32, (3,3,15), activation="relu", padding="same")(decoder_dense)
		decoder_conv_1 = layers.BatchNormalization()(decoder_conv_1)
		decoder_conv_2 = layers.Conv3DTranspose(16, (3,3,15), activation="relu", padding="same")(decoder_conv_1)
		decoder_conv_2 = layers.BatchNormalization()(decoder_conv_2)
		#decoder_conv_3 = layers.Conv3DTranspose(1, (3,3,15), activation="relu", padding="same")(decoder_conv_2)
		#decoder_conv_3 = layers.BatchNormalization()(decoder_conv_3)
		#decoder_conv_4 = layers.Conv3DTranspose(1, (2,2,15), activation="sigmoid", padding="same")(decoder_conv_3)
		#decoder_conv_4 = layers.BatchNormalization()(decoder_conv_4)
		#decoder_conv_5 = layers.Conv3DTranspose(16, (2,2,15), activation="relu", padding="same")(decoder_conv_1)
		#decoder_conv_5 = layers.BatchNormalization()(decoder_conv_5)
		#decoder_conv_6 = layers.Conv3DTranspose(8, (2,2,15), activation="relu", padding="same")(decoder_conv_5)
		#decoder_conv_6 = layers.BatchNormalization()(decoder_conv_6)
		#decoder_conv_7 = layers.Conv3DTranspose(1, (2,2,15), activation="sigmoid", padding="same")(decoder_conv_6)
		reconstructed_input = decoder_conv_2

		encoder = keras.Model(inputs=encoder_input, outputs=embedding, name='encoder')
		decoder = keras.Model(inputs=decoder_input, outputs=reconstructed_input, name='decoder')
		autoencoder = keras.Model(inputs=encoder_input, outputs=decoder(encoder(encoder_input)), name='autoencoder')

		return {'encoder': encoder, 'decoder': decoder, 'autoencoder': autoencoder}

	def _define_encoder(self) -> Model:
		return self._model_helper()['encoder']

	def _define_decoder(self):
		return self._model_helper()['decoder']

	def _define_model(self) -> Model:
		return self._model_helper()['autoencoder']


from functools import wraps
import numpy as np
import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import layers, Model

from chesspos.models.autoencoder import AutoencoderModel

class CnnAutoencoder(AutoencoderModel):
	@wraps(AutoencoderModel.__init__)
	def __init__(self, **kwargs):
		self._final_conv_shape = None
		self.embedding_size = 256
		super().__init__(**kwargs)

	def _model_helper(self) -> dict:
		encoder_input = layers.Input(shape=(8,8,15,1), dtype=tf.float16)

		# Encoder
		x = layers.Conv3D(32, (3, 3, 15), activation="relu", padding="same")(encoder_input)
		x = layers.MaxPooling3D((2, 2, 1), padding="same")(x)
		x = layers.Conv3D(32, (3, 3, 15), activation="relu", padding="same")(x)
		x = layers.MaxPooling3D((2, 2, 1), padding="same")(x)

		# Decoder
		decoder_input = layers.Input(shape=(2,2,15,32), dtype=tf.float16)
		y = layers.Conv3DTranspose(32, (3, 3, 15), strides=(2,2,1), activation="relu", padding="same")(decoder_input)
		y = layers.Conv3DTranspose(32, (3, 3, 15), strides=(2,2,1), activation="relu", padding="same")(y)
		y = layers.Conv3D(1, (8, 8, 15), activation="sigmoid", padding="same")(y)

		encoder = keras.Model(inputs=encoder_input, outputs=x, name='encoder')
		decoder = keras.Model(inputs=decoder_input, outputs=y, name='decoder')
		autoencoder = keras.Model(inputs=encoder_input, outputs=decoder(encoder(encoder_input)), name='autoencoder')

		return {'encoder': encoder, 'decoder': decoder, 'autoencoder': autoencoder}

	def _define_encoder(self) -> Model:
		return self._model_helper()['encoder']

	def _define_decoder(self):
		return self._model_helper()['decoder']

	def _define_model(self) -> Model:
		return self._model_helper()['autoencoder']


from abc import ABC, abstractmethod

import tensorflow as tf
from tensorflow import keras
from keras.models import load_model, Model

class SaveableModel(ABC):
	def __init__(self, save_dir: str, **kwargs) -> None:
		self.save_dir = save_dir
		self.model = self._define_model()
		for key, value in kwargs.items():
			print(f"Unknown argument: {key}={value} encountered and ignored for model instantiation.")

	@abstractmethod
	def _define_model(self) -> Model:
		pass

	def get_model(self) -> Model:
		return self.model

	def save(self) -> None:
		self.model.save(f"{self.save_dir}/model")


	def load(self) -> None:
		self.model = load_model(f"{self.save_dir}/model")


# Embedding models
from chesspos.models.autoencoder import AutoencoderModel
from chesspos.models.dense_autoencoder import DenseAutoencoder
from chesspos.models.resnet_autoencoder import ResnetAutoencoder
from chesspos.models.cnn_autoencoder import CnnAutoencoder



import chess
import numpy as np

def board_to_bitboard(board):
	embedding = np.array([], dtype=bool)
	for color in [1, 0]:
		for i in range(1, 7): # P N B R Q K / white
			bmp = np.zeros(shape=(64,)).astype(bool)
			for j in list(board.pieces(i, color)):
				bmp[j] = True
			embedding = np.concatenate((embedding, bmp))
	additional = np.array([
		bool(board.turn),
		board.has_kingside_castling_rights(chess.WHITE),
		board.has_queenside_castling_rights(chess.WHITE),
		board.has_kingside_castling_rights(chess.BLACK),
		board.has_queenside_castling_rights(chess.BLACK)
	])
	embedding = np.concatenate((embedding, additional))
	return embedding

def bitboard_to_board(bb):
	assert bb.shape == (773,)

	# set up empty board
	reconstructed_board = chess.Board()
	reconstructed_board.clear()
	# loop over all pieces and squares
	for color in [1, 0]: # white, black
		for i in range(1, 7): # P N B R Q K
			idx = (1-color)*6 + i - 1
			piece = chess.Piece(i,color)
			bitmask = bb[idx*64:(idx+1)*64]
			squares = np.flatnonzero(bitmask)
			for square in squares:
				reconstructed_board.set_piece_at(square,piece)

	# set turn
	reconstructed_board.turn = bb[768]

	# set castling rights 
	castling_rights = ''
	if bb[769]: castling_rights += 'Q'
	if bb[770]: castling_rights += 'K'
	if bb[771]: castling_rights += 'q'
	if bb[772]: castling_rights += 'k'
	reconstructed_board.set_castling_fen(castling_rights)

	return reconstructed_board

def board_to_tensor(board):
	embedding = np.empty((8,8,15), dtype=bool)
	# one plane per piece
	for color in [1, 0]:
		for i in range(1, 7): # P N B R Q K / white
			index = (1-color)*6 + i - 1
			bmp = np.zeros(shape=(64,)).astype(bool)
			for j in list(board.pieces(i, color)):
				bmp[j] = True
			bmp = bmp.reshape((8,8))
			embedding[:,:, index] = bmp

	# castling rights at plane embedding(:,:,12)
	embedding[0,0,12] = board.has_queenside_castling_rights(chess.WHITE)
	embedding[0,7,12] = board.has_kingside_castling_rights(chess.WHITE)
	embedding[7,0,12] = board.has_queenside_castling_rights(chess.BLACK)
	embedding[7,7,12] = board.has_kingside_castling_rights(chess.BLACK)

	# en passant squares at plane embedding(:,:,13)
	en_passant = np.zeros((64,), dtype=bool)
	if board.has_legal_en_passant():
		en_passant[board.ep_square] = True
	en_passant = en_passant.reshape((8,8))
	embedding[:,:,13] = en_passant

	# turn at plane embedding(:,:,14)
	embedding[0,0,14] = board.turn

	return embedding

def tensor_to_board(tensor, threshold=0.5):
	assert tensor.shape == (8,8,15), f"tensor_to_board encounterer an input with invalid shape {tensor.shape}, expected shape (8,8,15)"
	tensor = np.where(tensor > threshold, 1, 0)

	# set up empty board
	reconstructed_board = chess.Board()
	reconstructed_board.clear()

	# loop over all pieces and squares
	for color in [1, 0]: # white, black
		for i in range(1, 7): # P N B R Q K
			idx = (1-color)*6 + i - 1
			piece = chess.Piece(i,color)
			square_bitmask = tensor[:,:,idx].reshape((64,))
			squares = np.flatnonzero(square_bitmask)
			for square in squares:
				reconstructed_board.set_piece_at(square,piece)
	
	# set castling rights
	castling_rights = ''
	if tensor[0,0,12]: castling_rights += 'Q'
	if tensor[0,7,12]: castling_rights += 'K'
	if tensor[7,0,12]: castling_rights += 'q'
	if tensor[7,7,12]: castling_rights += 'k'
	reconstructed_board.set_castling_fen(castling_rights)

	# set en passant square
	en_passant = tensor[:,:,13].reshape((64,))
	if np.any(en_passant):
		reconstructed_board.ep_square = np.flatnonzero(en_passant)[0]

	# set turn
	reconstructed_board.turn = tensor[0,0,14]

	return reconstructed_board			



import logging
from typing import Callable, List
from attrs import define, field

import numpy as np
import h5py
import chess
import chess.pgn

from chesspos.utils.utils import correct_file_ending

@define
class PgnExtractor():
	pgn_path: str
	save_path: str
	game_processor: Callable[[chess.pgn.Game], np.ndarray]
	game_filter: Callable[[chess.pgn.Headers], bool] = lambda header: False
	chunk_size: int = 100000
	log_level: int = logging.ERROR
	_logger: logging.Logger = field(init=False)
	_game_counter: int = field(init=False, default=0)
	_chunk_counter: int = field(init=False, default=0)
	_encoding_counter: int = field(init=False, default=0)
	_encoding_shape: np.ndarray = field(init=False)
	_encoding_type: np.dtype = field(init=False)
	_discarded_games: int = 0
	_processed_games: int = 0
	
	@_logger.default
	def _get_default_logger(self):
		logger = logging.getLogger(__name__)
		logging.basicConfig(
			level=self.log_level,
			format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
			filename="pgn_extract.log")
		return logger
	@_encoding_shape.default
	def _get_encoding_shape(self):
		_, shape = self._get_encoding_type_and_shape()
		return shape

	@_encoding_type.default
	def _get_encoding_type(self):
		dtype, _ = self._get_encoding_type_and_shape()
		return dtype 

	def _get_encoding_type_and_shape(self):
		with open(correct_file_ending(self.pgn_path, "pgn"), 'r') as f:
			game = chess.pgn.read_game(f)
			encoding = self.game_processor(game)
			self._logger.info(f"Type of encoding: {encoding.dtype}, shape of encoding: {encoding.shape[1:]}")
			return encoding.dtype, encoding.shape[1:]

	def _write_chunk_to_file(self, chunk: np.ndarray, metadata: np.ndarray):
		self._logger.info(f"Saving chunk {self._chunk_counter}")
		fname = correct_file_ending(self.save_path, "h5")

		try:
			with h5py.File(fname, "a") as save_file:
				data1 = save_file.create_dataset(f"encoding_{self._chunk_counter}", data=chunk, compression="gzip", compression_opts=9)
				data2 = save_file.create_dataset(f"game_id_{self._chunk_counter}", data=metadata, compression="gzip", compression_opts=9)
				self._logger.info(f"Saved encodings with shape {chunk.shape}")
		except Exception as e:
			self._logger.error(f"Could not save chunk {self._chunk_counter}", exc_info=True)
			raise e

		self._chunk_counter += 1
		self._encoding_counter = 0

	def _games(self, number_games):
		with open(correct_file_ending(self.pgn_path, "pgn"), 'r') as pgn_file:
			while True:
				header = chess.pgn.read_headers(pgn_file)
				self._game_counter += 1

				# Get next suitable game or finish extraction
				if header is None:
					self._logger.info(f"Processed {self._game_counter} games in total")
					yield None
				elif self.game_filter(header):
					self._discarded_games += 1
					continue
				elif self._game_counter >= number_games:
					self._logger.info(f"Processed {self._game_counter} games in total")
					yield None
				else:
					self._processed_games += 1
					yield chess.pgn.read_game(pgn_file)

	def extract(self, number_games: int = int(1e18)):
		encoding_chunk = np.empty((self.chunk_size, *self._encoding_shape), dtype=self._encoding_type)
		game_id = np.empty((self.chunk_size,), dtype=np.int32)

		for game in self._games(number_games):
			if game is None:
				chunk = encoding_chunk[:self._encoding_counter]
				metadata = game_id[:self._encoding_counter]
				self._write_chunk_to_file(chunk, metadata)
				break

			# Extract information from that game
			encodings = self.game_processor(game)
			# Edge case: chunksize reached
			number_encodings = min(encodings.shape[0], self.chunk_size - self._encoding_counter)
			#self._logger.info(f"Extracted {number_encodings} encodings from game {self._game_counter}")
			new_encoding_counter = self._encoding_counter + number_encodings
			encoding_chunk[self._encoding_counter:new_encoding_counter, ...] = encodings[:number_encodings, ...]
			game_id[self._encoding_counter:new_encoding_counter] = self._game_counter*np.ones(number_encodings, dtype=np.int32)
			self._encoding_counter = new_encoding_counter

			# Save chunk if it is full
			if self._encoding_counter == self.chunk_size:
				self._logger.info(f"Processed {self._processed_games} games in this chunk")
				self._logger.info(f"Filtered {self._discarded_games} games in this chunk")
				self._processed_games = 0
				self._discarded_games = 0
				self._write_chunk_to_file(encoding_chunk, game_id)
		
		# Log file headers
		fname = correct_file_ending(self.save_path, "h5")
		with h5py.File(fname, 'r') as hf:
			self._logger.info(f"Keys: {hf.keys()}")
			self._logger.info(f"Shape of encoding: {hf['encoding_0'].shape}")
			self._logger.info(f"Shape of game_id: {hf['game_id_0'].shape}")


from chesspos.preprocessing.sample_generator import SampleGenerator
from chesspos.preprocessing.pgn_extractor import PgnExtractor
from chesspos.preprocessing.game_processors import *
from chesspos.preprocessing.game_filters import *
from chesspos.preprocessing.board_converter import *
from chesspos.preprocessing.utils import *



import tensorflow as tf
from tensorflow import keras


class TripletLossLayer(keras.layers.Layer):
	def __init__(self, alpha, **kwargs):
		self.alpha = alpha
		super(TripletLossLayer, self).__init__(**kwargs)

	def triplet_loss(self, inputs):
		anchor, positive, negative = inputs
		pos_dist = tf.reduce_sum(tf.square(anchor-positive), axis=-1)
		neg_dist = tf.reduce_sum(tf.square(anchor-negative), axis=-1)
		return tf.reduce_mean(tf.maximum(pos_dist - neg_dist + self.alpha, 0), axis=0)

	def call(self, inputs):
		loss = self.triplet_loss(inputs)
		self.add_loss(loss)
		return loss

class AutoencoderTripletLossLayer(keras.layers.Layer):
	def __init__(self, triplet_weight_ratio, **kwargs):
		self.triplet_weight_ratio = triplet_weight_ratio
		super(AutoencoderTripletLossLayer, self).__init__(**kwargs)

	def triplet_loss(self, inputs):
		return inputs[0]

	def autoencoder_loss(self, inputs):
		_, inp_a, inp_p, inp_n, out_a, out_p, out_n = inputs
		reconstruct_anc = tf.reduce_sum(tf.square(inp_a-out_a), axis=-1)
		reconstruct_pos = tf.reduce_sum(tf.square(inp_p-out_p), axis=-1)
		reconstruct_neg = tf.reduce_sum(tf.square(inp_n-out_n), axis=-1)
		return tf.multiply(1./3., tf.reduce_mean(reconstruct_anc, axis=0) +\
			tf.reduce_mean(reconstruct_pos, axis=0) + tf.reduce_mean(reconstruct_neg, axis=0)
		)

	def call(self, inputs):
		loss = self.triplet_weight_ratio * self.triplet_loss(inputs) + self.autoencoder_loss(inputs)
		self.add_loss(loss)
		return loss

def encoder_cnn(input_layer, name='cnn_encoder'):
	# prepare slicing layers
	slice_board_layer = keras.layers.Layer(lambda x: x[:,:-5])
	slice_metadata_layer = keras.layers.Layer(lambda x: x[:-5,:])

	# slice input
	embedding_board = slice_board_layer(input_layer)
	embedding_metadata = slice_metadata_layer(input_layer)

	# reshape board layer
	embedding_board = keras.layers.Reshape((None,8,8,12))(embedding_board)
	# convolution and pooling layers
	embedding_board = keras.layers.Conv2D(filters = 16, kernel_size = (3, 3), activation='relu', padding='valid')(embedding_board)
	embedding_board = keras.layers.MaxPooling2D(pool_size = (2, 2), padding='valid')(embedding_board)
	embedding_board = keras.layers.Conv2D(filters = 16, kernel_size = (3, 3), activation='relu', padding='valid')(embedding_board)
	embedding_board = keras.layers.MaxPooling2D(pool_size = (2, 2), padding='valid')(embedding_board)
	embedding_board = keras.layers.Flatten()(embedding_board)
	embedding = keras.layers.Concatenate(axis=1)([embedding_board, embedding_metadata])

	return embedding





def triplet_network(input_size, embedding_size, hidden_layers=None, alpha=0.2):
	# Input layers
	anchor_input = keras.layers.Input((input_size,), name="anchor_input", dtype=float)
	positive_input = keras.layers.Input((input_size,), name="positive_input", dtype=float)
	negative_input = keras.layers.Input((input_size,), name="negative_input", dtype=float)

	# Generate the encodings (feature vectors) for the three positions
	embedding = embedding_network(input_size, embedding_size, hidden_layers=hidden_layers)
	embedding.summary()

	# Embeddings for the three inputs
	embedding_a = embedding(anchor_input)
	embedding_p = embedding(positive_input)
	embedding_n = embedding(negative_input)

	# TripletLoss Layer, initialize and incorporate into network, tie embeddings together
	triplet_layer = TripletLossLayer(alpha=alpha, name='triplet_loss_layer')
	triplet_loss = triplet_layer([embedding_a, embedding_p, embedding_n])

	# Cast as tf model
	model_triplet = keras.models.Model(
		inputs=[anchor_input, positive_input, negative_input],
		outputs=[triplet_loss, embedding_a, embedding_p, embedding_n]
	)

	# Compile the model
	def mean_pred(y_true, y_pred): # pylint: disable=unused-argument,dangerous-default-value
		return print("hello")

	optimizer = keras.optimizers.Adam(lr=0.00006)
	model_triplet.compile(
		loss=None,
		optimizer=optimizer,
		metrics=[mean_pred] # call to any metric not working, why?
	)

	# Print model summary
	model_triplet.summary()

	return model_triplet

def triplet_autoencoder(input_size, embedding_size, hidden_layers=None,
	triplet_weight_ratio=1.0, alpha=0.2):
	# optimizer for all models
	optimizer = keras.optimizers.Adam(lr=0.00006)

	# Input layers
	anchor_input = keras.layers.Input((input_size,), name="anchor_input", dtype=float)
	positive_input = keras.layers.Input((input_size,), name="positive_input", dtype=float)
	negative_input = keras.layers.Input((input_size,), name="negative_input", dtype=float)

	# Generate the encodings (feature vectors) for the three positions
	encoder = embedding_network(input_size, embedding_size,
		hidden_layers=hidden_layers, name="encoder_network"
	)
	encoder.compile(loss='mse', optimizer=optimizer)
	encoder.summary()

	# Embeddings for the three inputs
	encoder_a = encoder(anchor_input)
	encoder_p = encoder(positive_input)
	encoder_n = encoder(negative_input)

	# TripletLoss Layer, initialize and incorporate into network, tie embeddings together
	triplet_layer = TripletLossLayer(alpha=alpha, name='triplet_loss_layer')
	triplet_loss = triplet_layer([encoder_a, encoder_p, encoder_n])

	# Initialise decoder
	decoder = embedding_network(embedding_size, input_size,
		hidden_layers=hidden_layers[::-1], name="decoder_network"
	)
	decoder.compile(loss='mse', optimizer=optimizer)
	decoder.summary()

	# decode embeddings
	decoder_a = decoder(encoder_a)
	decoder_p = decoder(encoder_p)
	decoder_n = decoder(encoder_n)

	# Autoencoder Loss Layer
	autoencoder_layer = AutoencoderTripletLossLayer(
		triplet_weight_ratio=triplet_weight_ratio,
		name="autoencoder_layer"
	)
	autoencoder_loss = autoencoder_layer(
		[triplet_loss, anchor_input, positive_input, negative_input, decoder_a, decoder_p, decoder_n]
	)

	# Cast as tf model
	model_autoencoder_triplet = keras.models.Model(
		inputs=[anchor_input, positive_input, negative_input],
		outputs=[autoencoder_loss, encoder_a, encoder_p, encoder_n]
	)
	model_autoencoder_triplet.summary()

	# Compile the model
	def mean_pred(y_true, y_pred): # pylint: disable=unused-argument,dangerous-default-value
		return print("hello")

	model_autoencoder_triplet.compile(
		loss=None,
		optimizer=optimizer,
		metrics=[mean_pred] # call to any metric not working, why?
	)

	return {'autoencoder': model_autoencoder_triplet, 'encoder': encoder, 'decoder': decoder}

def autoencoder(input_size, embedding_size, hidden_layers=None, hidden_decoder=None):

	optimizer = keras.optimizers.Adam(lr=0.0006)
	loss = 'mse'
	metrics = ['mse','binary_crossentropy']

	# input layer
	inp = keras.layers.Input((input_size,), name="input_layer", dtype=float)

	# generate encoder model
	encoder = embedding_network(input_size, embedding_size,
		hidden_layers=hidden_layers, name="encoder_network"
	)
	encoder.compile(loss='mse', optimizer=optimizer)
	encoder.summary()

	# instantiate encoder
	encoded = encoder(inp)

	# generate decoder model
	if hidden_decoder == None or len(hidden_decoder) == 0:
		hidden_decoder = hidden_layers[::-1]
	decoder = embedding_network(embedding_size, input_size,
		hidden_layers=hidden_decoder, name="decoder_network"
	)
	decoder.compile(loss='mse', optimizer=optimizer)
	decoder.summary()

	# instantiate decoder
	decoded = decoder(encoded)

	# generate autoencoder model
	autoencoder = keras.models.Model(
		inputs=inp,
		outputs=decoded
	)
	autoencoder.compile(
		optimizer=optimizer,
		loss=loss,
		metrics=metrics
	)
	autoencoder.summary()

	return {'autoencoder':autoencoder, 'encoder':encoder, 'decoder':decoder}



from os.path import abspath

import tensorflow as tf
import numpy as np
import matplotlib.pyplot as plt

class SkMetrics(tf.keras.callbacks.Callback):
	'''
	Custom callback to monitor triplet classification accuracy on epoch end.
	'''
	def __init__(self, valid_data, batch_size, steps_per_callback=10):
		super(SkMetrics, self).__init__()
		self.valid_data = valid_data
		self.batch_size = batch_size
		self.steps_per_callback = steps_per_callback

	def predict_correct(self, predictions):
		anchor = predictions[1]
		positive = predictions[2]
		negative = predictions[3]
		pos_dist = tf.reduce_sum(tf.square(anchor-positive), axis=-1)
		neg_dist = tf.reduce_sum(tf.square(anchor-negative), axis=-1)
		return tf.reduce_sum(tf.cast(pos_dist < neg_dist, dtype=tf.int32), axis=0)

	def on_train_begin(self, logs={}): # pylint: disable=unused-argument,dangerous-default-value
		self.frac_correct = [] # pylint: disable=attribute-defined-outside-init

	def on_epoch_end(self, epoch, logs={}): # pylint: disable=unused-argument,dangerous-default-value
		correct = tf.Variable(0, dtype=tf.int32)
		for i in range(self.steps_per_callback):
			predictions = self.model.predict_on_batch(next(self.valid_data))
			correct.assign_add(self.predict_correct(predictions))
		frac = tf.cast(correct, dtype=tf.float32)/tf.Variable(self.batch_size*self.steps_per_callback, dtype=tf.float32)
		self.frac_correct.append(frac.numpy())
		print(f" triplet_acc: {self.frac_correct[-1]}")

def plot_metrics(save_dir, loss_arr, loss_labels, other_metric=None, other_label=None):
	assert len(loss_arr) == len(loss_labels)

	fig, ax1 = plt.subplots()
	ax1.set_title('Training progress')
	ax1.set_xlabel('epoch')
	ax1.set_ylabel('loss')
	for (i, el) in enumerate(loss_arr):
		ax1.plot(np.arange(len(el)), el, 'x-', label=loss_labels[i])
	plt.legend()

	if other_metric is not None and other_label is not None:
		ax2 = ax1.twinx()  # instantiate a second axes that shares the same x-axis
		color = 'tab:red'
		ax2.set_ylabel(other_label, color=color)
		ax2.plot(np.arange(len(other_metric)), other_metric, 'ro-', label=other_label)
		ax2.tick_params(axis='y', labelcolor=color)
		plt.legend()

	fig.tight_layout()
	plt.savefig(save_dir+"/train_progress.png")

def save_metrics(metric_arrays, metric_labels, save_dir, plot=True):
	# get absolute path
	path = abspath(save_dir)
	# save metrics to file
	if len(metric_labels) == len(metric_arrays):
		# this is recommended
		d = dict(zip(metric_labels, metric_arrays))
		np.savez_compressed(f"{path}/metrics",**d)
	else:
		raise ValueError("No metric labels provided.")

	if plot:
		*loss, special = metric_arrays
		*label, special_label = metric_labels
		plot_metrics(path, loss, label, other_metric=special, other_label=special_label)



import h5py
import numpy as np

from chesspos.utils.utils import correct_file_ending

def samples_from_file(file, table_id_prefix, dtype=np.float32):
	'''
	Return samples from relevant table in a file.
	'''
	fname = correct_file_ending(file, 'h5')
	samples = []
	with h5py.File(fname, 'r') as hf:
		print(f"keys in {fname}: {hf.keys()}")
		for key in hf.keys():
			if table_id_prefix in key:
				samples.extend(hf[key][:])
	return np.asarray(samples, dtype=dtype)

def samples_from_file_array(files, table_id_prefix, dtype=np.float32):
	'''
	Return specified tuples from all relevant tables in a list of files.
	'''
	samples = None
	for i, file in enumerate(files):
		tmp_samples = samples_from_file(file, table_id_prefix, dtype=dtype)
		if i == 0:
			samples = np.empty((len(files), *tmp_samples.shape), dtype=dtype)
		samples[i, ...] = samples_from_file(file, table_id_prefix, dtype=dtype)

	return samples.reshape((-1, *samples.shape[2:]))

def samples_generator_from_file(file, table_id_prefix, dtype=np.float32):
	'''
	Return generator of samples from relevant table in a file.
	'''
	fname = correct_file_ending(file, 'h5')
	samples = []
	with h5py.File(fname, 'r') as hf:
		for key in hf.keys():
			if table_id_prefix in key:
				yield np.asarray(hf[key][:], dtype=dtype)

def sample_generator_from_file_array(files, table_id_prefix, dtype=np.float32):
	'''
	Yield generator of samples from all relevant tables in a list of files.
	'''
	for file in files:
		for sample in samples_generator_from_file(file, table_id_prefix, dtype=dtype):
			yield sample



from functools import wraps
from typing import Callable, Dict, List, overload
import os
import math
import pickle
import numpy as np
import matplotlib.pyplot as plt
import chess


import tensorflow as tf
from tensorflow import keras

from chesspos.models.saveable_model import SaveableModel
from chesspos.preprocessing.sample_generator import SampleGenerator

class TrainableModel(SaveableModel):
	@wraps(SaveableModel.__init__)
	def __init__(
		self,
		train_generator: SampleGenerator,
		test_generator: SampleGenerator,
		train_steps_per_epoch: int,
		test_steps_per_epoch: int,
		optimizer = 'rmsprop',
		board_to_input: Callable[[chess.Board], np.ndarray] = None,
		loss = None,
		metrics = None,
		hide_tf_warnings: bool = True,
		tf_callbacks = None,
		**kwargs
	) -> None:

		self.EXCESS_VALIDATION_EPOCHS = 10

		self.board_to_input = board_to_input
		self.optimizer = optimizer
		self.loss = loss
		self.metrics = metrics

		self.train_generator = train_generator
		self.test_generator = test_generator
		self.train_steps_per_epoch = train_steps_per_epoch
		self.test_steps_per_epoch = test_steps_per_epoch
		self.hide_tf_warnings = hide_tf_warnings

		super(TrainableModel, self).__init__(**kwargs)

		self.tf_callbacks = self._set_tf_callbacks(tf_callbacks)
		print(self.loss)

	def _set_tf_callbacks(self, callback_array) -> List[keras.callbacks.Callback]:
		callbacks = []
		if callback_array is None:
			pass
		elif isinstance(callback_array, list):
			for callback in callback_array:
				if isinstance(callback, (
					keras.callbacks.EarlyStopping,
					keras.callbacks.ModelCheckpoint,
					keras.callbacks.TensorBoard,
					keras.callbacks.Callback)
				):
					callbacks.append(callback)
				elif isinstance(callback, str):
					if callback == 'early_stopping':
						callbacks.append(
							keras.callbacks.EarlyStopping(monitor='val_loss', min_delta=0.02, patience=10,
							verbose=0, mode='min', restore_best_weights=True)
						)
					elif callback == 'checkpoints':
						save_dir = os.path.abspath(self.save_dir)
						callbacks.append(
							keras.callbacks.ModelCheckpoint(filepath=save_dir+"/checkpoints/cp-{epoch:04d}.ckpt",
							save_weights_only=False, save_best_only=True, mode='min', verbose=1)
						)
					else:
						print(f"WARNING: illegal argument {callback} in tf_callbacks, skipping.")
		else:
			raise ValueError("'callback_array' needs to be a list.")
		
		return callbacks


	def _train_samples(self) -> int:
		return  1.0 * self.train_generator.number_samples


	def _test_samples(self) -> int:
		return 1.0 * self.test_generator.number_samples


	def _train_epochs(self) -> int:
		return self._train_samples() / self.train_generator.batch_size / self.train_steps_per_epoch


	def _test_epochs(self) -> int:
		return self._test_samples() / self.test_generator.batch_size / self.test_steps_per_epoch


	def get_model(self) -> keras.Model:
		return self.model


	def _compile(self) -> None:
		self.model.compile(optimizer=self.optimizer, loss=self.loss, metrics=self.metrics)
		keras.utils.plot_model(self.model, to_file=f"{self.save_dir}/model.png", show_shapes=True)
		self.model.summary()


	def _check_train_test_ratio(self):
		train_epochs = self._train_epochs()
		test_epochs = self._test_epochs()
		print(f'You have enough training samples for {train_epochs} epochs and enough validation samples for {test_epochs} epochs.')

		if train_epochs > test_epochs:
			raise ValueError("Not enough validation samples provided to start training! Cancelling.")
		else:
			print(f"\nTraining on {self._train_samples() / 1.e6} million training samples.")
			print(f"Validating on {math.floor(train_epochs) * self.test_generator.batch_size * self.test_steps_per_epoch / 1.e6} million validation samples.")
			if test_epochs > self.EXCESS_VALIDATION_EPOCHS + train_epochs:
				print("WARNING: your are providing much more validation samples than necessary. Those could be used for training instead.")

	def _plot_train_history(self, history: keras.callbacks.History) -> None:
		# summarize history for loss
		plt.plot(history.history['loss'])
		plt.plot(history.history['val_loss'])
		plt.title('model loss')
		plt.ylabel('loss')
		plt.xlabel('epoch')
		plt.legend(['train', 'test'], loc='upper left')
		plt.savefig(self.save_dir+'/loss.png')

	def train(self) -> Dict:
		if self.hide_tf_warnings:
			os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'

		# TODO: benchmark
		#train_generator = self.train_generator.get_generator()
		#test_generator = self.test_generator.get_generator()
		train_generator = self.train_generator.get_tf_dataset()
		test_generator = self.test_generator.get_tf_dataset()

		self._check_train_test_ratio()

		history = self.model.fit(
			train_generator,
			steps_per_epoch = self.train_steps_per_epoch,
			epochs = math.floor(self._train_epochs()),
			validation_data = test_generator,
			validation_steps = self.test_steps_per_epoch,
			callbacks = self.tf_callbacks
		)
		return history

	def predict(self, samples: np.ndarray) -> np.ndarray:
		return self.model.predict(samples)

	def _check_input_converter(self) -> None:
		if self.board_to_input is None:
			raise ValueError("No board to model input converter set.")

		start_board = chess.Board()
		converter_shape = self.board_to_input(start_board).shape
		input_shape = self.train_generator.sample_shape
		if not converter_shape == input_shape:
			raise ValueError(f"board_to_input converter is not compatible with model input shape.\n"
				f"Expected shape: {converter_shape}\n"
				f"Input shape: {input_shape}")

	def predict_from_board(self, boards: List[chess.Board]) -> np.ndarray:
		self._check_input_converter()
		inputs = np.empty((len(boards), *self.train_generator.sample_shape))
		for i, board in enumerate(boards):
			inputs[i] = self.board_to_input(board)
		return self.predict(inputs)

	def evaluate(self, samples: np.ndarray, labels: np.ndarray) -> np.float32:
		return self.model.evaluate(samples, labels)

	def evaluate_from_board(self, boards: List[chess.Board], labels) -> np.float32:
		self._check_input_converter()
		inputs = np.empty((len(boards), *self.train_generator.sample_shape))
		for i, board in enumerate(boards):
			inputs[i] = self.board_to_input(board)
		return self.evaluate(inputs, labels)



import os
import h5py
from typing import Callable, List, Tuple
import numpy as np
import tensorflow as tf

from chesspos.utils.utils import correct_file_ending, files_from_directory
from chesspos.preprocessing.utils import sample_generator_from_file_array


class SampleGenerator():
	def __init__(
		self,
		sample_dir,
		sample_preprocessor: Callable[[np.ndarray], np.ndarray],
		batch_size=16,
		sample_type=np.float32,
	):
		self.H5_COL_KEY = 'encoding'
		self.sample_dir = sample_dir
		self.sample_preprocessor = sample_preprocessor
		self.batch_size = batch_size
		self.sample_type = sample_type
		self.number_samples, self.sample_shape = self._get_sample_dimensions()
		self.generator_function = self._construct_generator_function()


	def _construct_generator_function(self):
		def generator_function():
			assert(self.sample_shape is not None, "SampleGenerator has not been initialized with a sample shape.")
			sample_files = files_from_directory(os.path.abspath(self.sample_dir), file_type="h5")
			for samples in sample_generator_from_file_array(sample_files, self.H5_COL_KEY, self.sample_type):
				tmp_samples = samples
				i = 0
				while len(tmp_samples) >= self.batch_size:
					yield self.sample_preprocessor(samples[:self.batch_size, ...])
					samples = samples[self.batch_size:, ...]
		return generator_function


	def _get_generator_signature(self):
		sample = next(self.generator_function())
		out_shape = None
		if isinstance(sample, np.ndarray):
			out_shape = sample.shape
		elif isinstance(sample, (Tuple, List)):
			out_shape = []
			for s in sample:
				out_shape.append(s.shape)
		return out_shape

	def get_tf_dataset(self):
		out_shape = self._get_generator_signature()
		output_signature = []
		for s in out_shape:
			output_signature.append(tf.TensorSpec(s, self.sample_type))

		return tf.data.Dataset.from_generator(
			self.generator_function,
			output_signature=tuple(output_signature)
		)


	def get_generator(self):
		return self.generator_function()


	def _get_sample_dimensions(self):
		"""
		Return the number of samples (first dimension) and the shape of the samples (other dimensions).
		"""
		samples = 0
		shape = None
		sample_files = files_from_directory(os.path.abspath(self.sample_dir), file_type="h5")
		for i, file in enumerate(sample_files):
			fname = correct_file_ending(file, 'h5')
			with h5py.File(fname, 'r') as hf:
				for key in hf.keys():
					if self.H5_COL_KEY in key:
						samples += hf[key].shape[0]
						if i == 0:
							shape = hf[key].shape[1:]
						else:
							assert(shape == hf[key].shape[1:], "Shape of samples in file {} does not match shape of samples in file {}".format(i, i-1))
		return samples, shape



from os import listdir
from os.path import isfile, join, abspath

def files_from_directory(directory, file_type=None):
	absdir = abspath(directory)
	file_arr = None
	if file_type is None:
		file_arr = [join(absdir, f) for f in listdir(absdir) if isfile(join(absdir, f))]
	else:
		file_arr = [join(absdir, f) for f in listdir(absdir)
								if (isfile(join(absdir, f)) and f.endswith(file_type))]
	return file_arr

def correct_file_ending(file, ending):
	len_ending = len(ending)
	out_file = ""
	if file[-len_ending:] == ending:
		out_file = file
	else:
		out_file = f"{file}.{ending}"
	return out_file



import functools
import time

def count_calls(func):
    @functools.wraps(func)
    def wrapper_count_calls(*args, **kwargs):
        wrapper_count_calls.num_calls += 1
        print(f"Call {wrapper_count_calls.num_calls} of {func.__name__!r}")
        return func(*args, **kwargs)
    wrapper_count_calls.num_calls = 0
    return wrapper_count_calls
	
def timer(func):
    """Print the runtime of the decorated function"""
    @functools.wraps(func)
    def wrapper_timer(*args, **kwargs):
        start_time = time.perf_counter()
        value = func(*args, **kwargs)
        end_time = time.perf_counter()
        run_time = end_time - start_time
        print(f"Finished {func.__name__!r} in {run_time:.4f} secs")
        return value
    return wrapper_timer



from setuptools import setup, find_packages

setup(
	name='chesspos',
	version='0.1.3',
	description='A library for manipulating, learning and searching chess positions',
	url='https://github.com/patrickfrank1/chesspos/',
	author='Patrick Frank',
	author_email='patrickfrank1@outlook.com',
	license='GPLv3',
	packages=find_packages(),
	classifiers=[
		'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
		'Operating System :: OS Independent',
		'Programming Language :: Python :: 3',
		'Topic :: Scientific/Engineering :: Artificial Intelligence',
		'Topic :: Software Development :: Pre-processors',
		'Topic :: Games/Entertainment :: Board Games'
	],
	zip_safe=False,
	python_requires='>=3.7',
	install_requires=[
		'chess',
		'colorama',
		'faiss',
		'h5py',
		'keras',
		'matplotlib',
		'numpy',
		'scikit_learn',
		'setuptools',
		'tensorflow',
		'pyyaml'
	]
)



from abc import abstractmethod
from functools import cmp_to_key, wraps
from typing import Callable, List, overload
from colorama import Fore, Style
import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import layers, Model

import numpy as np
import chess

from chesspos.models.trainable_model import TrainableModel

class AutoencoderModel(TrainableModel):
	@wraps(TrainableModel.__init__)
	def __init__(
		self,
		output_to_board: Callable[[np.ndarray], chess.Board] = None,
		**kwargs
	) -> None:
		super(AutoencoderModel, self).__init__(**kwargs)

		self.encoder = self._define_encoder()
		self.decoder = self._define_decoder()
		self.output_to_board = output_to_board

		self._compile()

	def _compile(self) -> None:
		super()._compile()
		self.encoder.compile(optimizer=self.optimizer, loss=None, metrics=None)
		keras.utils.plot_model(self.encoder, to_file=f"{self.save_dir}/encoder.png", show_shapes=True)

		self.decoder.compile(optimizer=self.optimizer, loss=None, metrics=None)
		keras.utils.plot_model(self.decoder, to_file=f"{self.save_dir}/decoder.png", show_shapes=True)

	@abstractmethod
	def _define_encoder(self) -> Model:
		pass

	@abstractmethod
	def _define_decoder(self) -> Model:
		pass

	def _define_model(self) -> Model:
		encoder = self._define_encoder()
		decoder = self._define_decoder()
		autoencoder = decoder(encoder)
		return keras.Model(inputs=encoder.get_layer(index=0).input, outputs=autoencoder, name="autoencoder")

	def get_encoder(self):
		return self.encoder


	def get_decoder(self):
		return self.decoder

	@staticmethod
	def binarize_array(array, threshold=0.5):
		return np.where(array > threshold, True, False)

	@staticmethod
	def print_board(board: chess.Board) -> None:
		board = board.__str__().split("\n")
		output = ""
		output += ''.join([f"{board[i]}\n" for i in range(len(board))]) +'\n'
		print(output)

	def _check_output_converter(self) -> None:
		if self.output_to_board is None:
			raise ValueError("No model output to board converter set.")

	def _predict_from_embedding(self, embedding: np.ndarray) -> np.ndarray:
		return self.decoder.predict(embedding)
	
	def predict_board_from_embedding(self, embedding: np.ndarray) -> chess.Board:
		return self.output_to_board(self._predict_from_embedding(embedding))

	def evaluate(self, samples: np.ndarray) -> np.float32:
		return super().evaluate(samples, samples)

	def evaluate_from_board(self, boards: List[chess.Board]) -> np.float32:
		self._check_input_converter()
		inputs = np.empty((len(boards), *self.train_generator.sample_shape))
		for i, board in enumerate(boards):
			inputs[i] = self.board_to_input(board)
		return self.evaluate(inputs)

	def _get_sorted_losses(self, sort_fn: Callable, number_samples: int = 1000) -> List[dict]:
		test_generator = self.test_generator.get_generator()
		batch_size = self.test_generator.batch_size
		batches = max(number_samples // batch_size, 1)
		samples = []

		for _ in range(batches):
			x, __ = next(test_generator)
			print(x.shape)
			for j in range(batch_size):
				input = x[j].reshape(1, *x[j].shape)
				loss = self.model.evaluate(input, input, verbose=0)
				samples.append({'input':input, 'loss':loss})

			sort_wrapper = lambda a, b: sort_fn(a['loss'], b['loss'])
			samples.sort(key=cmp_to_key(sort_wrapper))

		return samples

	def plot_best_samples(self, number_samples: int)-> None:
		max_loss = lambda x, y : x - y
		best_samples = self._get_sorted_losses(max_loss, number_samples=10*number_samples)[:number_samples]
		examples_out = ""
		print("Best reconstruction examples:")
		for i in range(number_samples):
			best_sample_input = best_samples[i]['input']
			examples_out += str(best_samples[i]['loss']) + '\n\n'
			examples_out += self._compare_input_to_prediction(best_sample_input)
		print(examples_out)

	def plot_worst_samples(self, number_samples: int)-> None:
		min_loss = lambda x, y : y - x
		worst_samples = self._get_sorted_losses(min_loss, number_samples=10*number_samples)[:number_samples]
		examples_out = ""
		print("Worst reconstruction examples:")
		for i in range(number_samples):
			worst_sample_input = worst_samples[i]['input']
			examples_out += str(worst_samples[i]['loss']) + '\n'
			examples_out += self._compare_input_to_prediction(worst_sample_input)
		print(examples_out)

	def _compare_input_to_prediction(self, input: np.ndarray) -> None:
		prediction = self.model.predict(input)
		prediction_board = self.output_to_board(prediction)
		prediction_board = prediction_board.__str__().split("\n")

		input_board = self.output_to_board(input)
		input_board = input_board.__str__().split("\n")

		for i in range(len(prediction_board)):
			prediction_board[i] = ''.join([f'{Fore.RED}{prediction_board[i][j]}{Style.RESET_ALL}' if input_board[i][j] != prediction_board[i][j] else prediction_board[i][j] for j in range(len(prediction_board[i]))])

		output_str = "Original:          Reconstructed:\n"
		output_str += ''.join([f"{input_board[i]}    {prediction_board[i]}\n" for i in range(len(input_board))]) +'\n'
		return output_str

	def interpolate(self, sample1: np.ndarray, sample2: np.ndarray) -> None:
		encoder = self.get_encoder()
		decoder = self.get_decoder()
		encoding1 = encoder.predict(sample1)
		encoding2 = encoder.predict(sample2)
		print(self.output_to_board(sample1).__str__())
		direction = encoding2 - encoding1
		for i in range(11):
			interpolation = decoder.predict(encoding1 + direction * i / 10.0)
			board = self.output_to_board(interpolation)
			print(board.__str__())
			print("\n")

	def interpolate2d(self, sample1, sample2, sample3, sample4):
		pass